import logging
import datetime
import decimal
import types
import funcy as fn
import json
from webob import Response, exc

import pibe
from pibe import JSONRouter

from .settings import settings

logger = logging.getLogger(__name__)

__all__ = (
    "http",
    "bad_request",
    "unauthorized",
    "forbidden",
    "not_found",
    "not_acceptable",
    "unprocessable_entity",
    "expectation_failed",
    "is_json",
    "no_content",
    "created",
)

class ExtRouter(JSONRouter):
    initialize_fns = []
    wsgi_middleware_fns = []

    def initialize(self):
        def func_decorator(func):
            self.initialize_fns.append(func)
            return func
        return func_decorator

    def wsgi_middleware(self):
        def func_decorator(func):
            self.wsgi_middleware_fns.append(func)
            return func
        return func_decorator

    def make_app(self, **opts):
        if opts.get("initialize", True) == True:
            settings.initialize(**opts)

            for init_func in self.initialize_fns:
                init_func(**opts)

        _app = self.application

        if opts.get("install_middleware", True) == True:
            for mw_fn in self.wsgi_middleware_fns:
                _app = mw_fn(_app, **opts)

        return _app

http = ExtRouter()
pibe.regex_fn["shortuuid"] = r"[2-9A-HJ-NP-Za-km-z]{22}"


def _raise_exc(
    exc_class,
    _default_error="Unknown Error Description",
    errors=None,
    error=None,
):
    raise exc_class(
        json= {"errors": errors or {"__all__": [error or _default_error]}},
        content_type="application/json",
    )


def bad_request(**kwargs):
    _raise_exc(exc.HTTPBadRequest, _default_error="Bad Request", **kwargs)


def unauthorized(**kwargs):
    _raise_exc(exc.HTTPUnauthorized, _default_error="Unauthorized", **kwargs)


def forbidden(**kwargs):
    _raise_exc(exc.HTTPForbidden, _default_error="Forbidden", **kwargs)


def not_found(**kwargs):
    _raise_exc(exc.HTTPNotFound, _default_error="Not Found", **kwargs)


def not_acceptable(**kwargs):
    _raise_exc(exc.HTTPNotAcceptable, _default_error="Not Acceptable", **kwargs)


def unprocessable_entity(**kwargs):
    _raise_exc(
        exc.HTTPUnprocessableEntity, _default_error="Unprocessable Entity", **kwargs
    )

def expectation_failed(**kwargs):
    _raise_exc(exc.HTTPExpectationFailed, _default_error="Expectation Failed", **kwargs)


@fn.decorator
def is_json(call):
    try:
        call.req.json
    except:
        not_acceptable(error="Invalid JSON request")
    return call()


@fn.decorator
def no_content(call):
    call()
    return Response(status=204)


@fn.decorator
def created(call):
    call()
    return Response(status=201)



import os
import pibe
import markdown
import textwrap
from jinja2 import Environment, BaseLoader
from pathlib import Path

MARKDOWN_CSS = """
.markdown-body p {margin-bottom: 10px;}
.markdown-body ul {margin-bottom: 10px;}
.markdown-body h1 {font-size: 150%; margin-bottom: 20px;}
.markdown-body h2 {font-size: 140%; margin-bottom: 15px;}
.markdown-body h3 {font-size: 130%; margin-bottom: 15px;}
.markdown-body h4 {font-size: 110%; margin-bottom: 15px;}
.markdown-body pre {margin-top: 10px; margin-bottom: 10px;}
.markdown-body ul {
    list-style-type: circle;
        margin-left: 20px;
        padding: 0;
}

._markdown-body pre code {
    background: #f8f8f8;
}

.markdown-body code {
	color: #E74C3C;
    background: #fff;
    border: 1px solid #e1e4e5;
    font-size: 75%;
    padding: 2px;
}

.markdown-body pre {
	color: #E74C3C;
    background: #fff;
    border: 1px solid #e1e4e5;
    font-size: 75%;
    padding: 2px;
}

.markdown-body pre code {
    border: 0px ;
}

.markdown-body .admonition {
    padding: 5px;
    border: 1px solid;
}
.markdown-body .admonition p {
    margin-bottom: 0px;
}

.markdown-body .admonition .admonition-title {
    font-size: 110%;
    font-weight: bold;
}

.markdown-body .admonition.danger {
    background-color: #fca5a5;
    color: #7f1d1d;
}

.markdown-body .admonition.warning {
    background-color: #fed7aa;
    color: #9a3412;
}

.markdown-body .admonition.info {
    background-color: #60a5fa;
    color: #1e3a8a;
}

"""


DOC_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <link href="https://cdn.jsdelivr.net/npm/tailwindcss@2.2.19/dist/tailwind.min.css" rel="stylesheet">
    <link rel="preconnect" href="https://fonts.googleapis.com">
<link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
<link href="https://fonts.googleapis.com/css2?family=Montserrat:ital,wght@0,100;0,200;0,300;0,400;0,500;0,600;0,700;0,800;0,900;1,100;1,200;1,300;1,400;1,500;1,600;1,700;1,800;1,900&display=swap" rel="stylesheet">

    <title>{{ title }}</title>
    <style>
        body {
            font-family: 'Montserrat', sans-serif;
        }

        {{ MARKDOWN_CSS }}
    </style>

</head>
<body class="bg-gray-50 leading-normal tracking-normal">
    <!-- Content -->
    <div class="container mx-auto mt-4 max-w-6xl">
        <div class="flex flex-row">
            <div class="flex-1 mb-10">
                <h1 class="text-4xl">{{ title }}</h1>
            </div>
            <div class="flex-0">
                <div class="flex flex-row text-4xl gap-2">
                    <div id="collapse-all" class="flex-1 border border-3 rounded-sm px-3 cursor-pointer">
                        &uarr;
                    </div>
                    <div id="expand-all" class="flex-1 border border-3 rounded-sm px-3 cursor-pointer">
                        &darr;
                    </div>

                </div>
            </div>
        </div>
        {% if description %}
        <div class="markdown-body mb-5">{{ description }}</div>
        {% endif %}
        {% for endpoint in endpoints %}
        <div class="border mb-5 p-2 ">
            <div class="flex flex-col md:flex-row ">
                <div class="flex-0 font-bold">
                    {% if endpoint.methods|length == 1 %}
                        {% if endpoint.methods[0] == "GET" %}
                            {% set method_classes = "bg-green-100 text-green-800" %}
                        {% elif endpoint.methods[0] == "POST" %}
                            {% set method_classes = "bg-blue-100 text-blue-800" %}
                        {% elif endpoint.methods[0] == "PUT" %}
                            {% set method_classes = "bg-blue-200 text-blue-500" %}
                        {% elif endpoint.methods[0] == "DELETE" %}
                            {% set method_classes = "bg-red-100 text-red-800" %}
                        {% elif endpoint.methods[0] == "HEAD" %}
                            {% set method_classes = "bg-blue-100 text-blue-800" %}
                        {% elif endpoint.methods[0] == "PATCH" %}
                            {% set method_classes = "bg-blue-100 text-blue-800" %}
                        {% elif endpoint.methods[0] == "OPTIONS" %}
                            {% set method_classes = "bg-blue-100 text-blue-800" %}
                        {% else %}
                            {% set method_classes = "bg-gray-200 text-gray-800" %}
                        {% endif %}
                    {% else %}
                        {% set method_classes = "bg-gray-200 text-gray-600" %}
                    {% endif %}
                    <div class="text-xl w-32 text-center p-2 border rounded-md {{ method_classes }}">{{ ", ".join(endpoint.methods)}}</div>
                </div>

                <div class="flex-1 py-2 px-5"><code class="font-bold">{{ endpoint.pattern|escape }}</code> <em>{{ endpoint.title }}</em></div>
                <div class="flex-0 py-2 flex flex-row text-4xl gap-2">
                    <div id="expand-{{ loop.index }}" class="flex-1 border border-3 rounded-sm px-3 cursor-pointer">
                        &darr;
                    </div>
                    <div id="collapse-{{loop.index}}" class="flex-1 border border-3 rounded-sm px-3 cursor-pointer hidden">
                        &uarr;
                    </div>
                </div>
            </div>
            <div id="endpoint-documentation-{{loop.index}}" class="markdown-body endpoint-documentation mt-5 mx-1 hidden">{{ endpoint.html_doc }}</div>
        </div>
        {% endfor %}
    </div>
    <script>
    {% for i in range(endpoints|length) %}
    document.getElementById("expand-{{ i+ 1 }}").addEventListener("click", (event) => {
        document.getElementById("expand-{{ i + 1 }}").classList.add("hidden");
        document.getElementById("collapse-{{ i + 1 }}").classList.remove("hidden");
        document.getElementById("endpoint-documentation-{{ i + 1 }}").classList.remove("hidden");
    });
    document.getElementById("collapse-{{ i+ 1 }}").addEventListener("click", (event) => {
        document.getElementById("expand-{{ i + 1 }}").classList.remove("hidden");
        document.getElementById("collapse-{{ i + 1 }}").classList.add("hidden");
        document.getElementById("endpoint-documentation-{{ i + 1 }}").classList.add("hidden");
    });
    {% endfor %}

    document.getElementById("expand-all").addEventListener("click", (event) => {
        {% for i in range(endpoints|length) %}
        document.getElementById("expand-{{ i + 1 }}").classList.add("hidden");
        document.getElementById("collapse-{{ i + 1 }}").classList.remove("hidden");
        document.getElementById("endpoint-documentation-{{ i + 1 }}").classList.remove("hidden");
        {% endfor %}
    });

    document.getElementById("collapse-all").addEventListener("click", (event) => {
        {% for i in range(endpoints|length) %}
        document.getElementById("expand-{{ i + 1 }}").classList.remove("hidden");
        document.getElementById("collapse-{{ i + 1 }}").classList.add("hidden");
        document.getElementById("endpoint-documentation-{{ i + 1 }}").classList.add("hidden");
        {% endfor %}
    });
    </script>
</body>
</html>
"""



def generate_documentation(router, **kwargs):
    endpoints = []
    for (regex, resource, methods, pattern, opts) in router:

        mrkdn_txt = ((textwrap.dedent(resource.__doc__) if resource.__doc__ else None)
                      or opts.get("doc")
                      or "*No documentation*")
        endpoints.append({
            "pattern": pattern,
            "methods": methods,
            "title": opts.get("title") or resource.__name__.replace("_", " ").capitalize(),
            "html_doc": markdown.markdown(mrkdn_txt, extensions=['admonition', 'fenced_code'])
        })
    title = kwargs.get("title") or "Documentation"
    description = (markdown.markdown(kwargs.get("description"))
                   if kwargs.get("description")
                   else None)
    template = Environment(loader=BaseLoader()).from_string(DOC_TEMPLATE).render(
        endpoints=endpoints,
        title=title,
        description=description,
        MARKDOWN_CSS=MARKDOWN_CSS
    )

    basepath = Path(kwargs.get("basepath", "docs_html"))

    if not basepath.exists():
        os.makedirs(basepath.absolute(), exist_ok=True)

    if not basepath.is_dir():
        raise ValueError("basepath is not a directory")

    for _extra_filepath in kwargs.get("extra", []):
        fp = Path(_extra_filepath)
        if not fp.exists():
            raise ValueError(f"file {fp} does not exist")
        with open(fp, 'r') as f:
            html = Environment(loader=BaseLoader()).from_string(EXTRA_TEMPLATE).render(
                content=markdown.markdown(f.read(), extensions=['admonition', 'fenced_code']),
                MARKDOWN_CSS=MARKDOWN_CSS
            )

        with open(basepath / f"{fp.stem}.html", 'w') as f:
            f.write(html)

    filepath = basepath / kwargs.get("filename", "index.html")

    with open(filepath, "w") as f:
        # Writing data to file
        f.write(template)



import math
import logging
import datetime
import decimal
import types
import funcy as fn

import json
import peewee as pw

from webob import Response, exc
from playhouse.shortcuts import model_to_dict, dict_to_model

from .http import *
from .db import *

__all__ = (
    "model_serializer",
    "get_object_or_404",
    "get_object_or_400",
    "get_object_or_422",
    "paginated",
    "filtered",
    "ordered",
    "skimmed",
    "object_list",
    "object_detail",
)

def model_serializer(
    model_class, fields=None, follow_m2m=True, exclude=None, extra=None, **serializers
):
    fields = fields or model_class._meta.fields.keys()
    if extra:
        fields = fields | extra

    if follow_m2m:
        fields = fields | model_class._meta.manytomany.keys()

    if exclude:
        fields = fields - set(exclude)

    def get_field_value(obj, field_name):
        if field_name in model_class._meta.fields:
            field = model_class._meta.fields[field_name]
        elif field_name in model_class._meta.manytomany:
            field = model_class._meta.manytomany[field_name]
        else:
            field = None

        if field and getattr(obj, field_name) is not None:
            if field.__class__ in (pw.DateField, pw.DateTimeField):
                return getattr(obj, field_name).isoformat()
            elif field.__class__ == pw.ForeignKeyField:
                return model_serializer(field.rel_model, follow_m2m=False)(
                    getattr(obj, field_name)
                )
            elif field.__class__ == pw.ManyToManyField:
                return model_serializer(field.rel_model, follow_m2m=False)(
                    getattr(obj, field_name)
                )
            elif field.__class__ == pw.DecimalField:
                return str(getattr(obj, field_name))

        return getattr(obj, field_name)

    defaults = dict(
        fn.lmap(
            lambda field_name: (
                field_name,
                lambda obj: get_field_value(obj, field_name),
            ),
            fields,
        )
    )
    serializers = fn.merge(defaults, serializers)

    def object_serializer(obj, project=None, omit=None):
        fields = serializers.keys()
        if project:
            fields = fields & project
        if omit:
            fields = fields - omit
        return dict([(f, serializers[f](obj)) for f in fields])

    def inner(qs, project=None, omit=None, **_fields):
        if isinstance(qs, pw.Model):
            return object_serializer(qs, project=project, omit=omit)
        return [object_serializer(obj, project=project, omit=omit) for obj in qs]

    return inner


def get_object_or_404(model_class, **kwargs):
    obj = model_class.get_or_none(**kwargs)
    if not obj:
        not_found(error="{} Not found".format(model_class.__name__))
    return obj


def get_object_or_400(model_class, **kwargs):
    obj = model_class.get_or_none(**kwargs)
    if not obj:
        bad_request(error="{} Not found".format(model_class.__name__))
    return obj


def get_object_or_422(model_class, **kwargs):
    obj = model_class.get_or_none(**kwargs)
    if not obj:
        unprocessable_entity(error="{} Not found".format(model_class.__name__))
    return obj


def paginated(req, qs, paginate_by=15, max_paginate_by=100):
    page = int(req.params.get("page", 1))
    paginate_by = min(int(req.params.get("paginate_by", paginate_by)), max_paginate_by)
    record_count = qs.count()
    page_count = int(math.ceil(record_count / paginate_by))
    return (
        qs.paginate(page, paginate_by),
        {
            "is_paginated": record_count > paginate_by,
            "page": page,
            "page_count": page_count,
            "record_count": record_count,
        },
    )


VALUE_CONVERSION = {"true": True, "false": False, "none": None}


def filtered(req, qs, fields=None, **filter_fns):
    params = req.params.mixed()
    filters = dict(
        fn.lmap(
            lambda x: (fn.cut_prefix(x[0], "filter__"), x[1]),
            fn.filter(lambda x: x[0].startswith("filter__"), fn.iteritems(params)),
        )
    )
    if filters:
        fn.walk_values(
            lambda v: VALUE_CONVERSION.get(v, v) if fn.isa(str)(v) else v, filters
        )
        qs = qs.filter(**fn.omit(filters, filter_fns.keys()))
        for key in fn.project(filters, filter_fns.keys()):
            qs = filter_fns[key](req, qs, filters[key])
    return qs


def ordered(req, qs, **order_fns):
    for order in req.params.getall("order_by"):
        direction = "desc" if order[0] == "-" else "asc"
        field_name = order[1:] if order[0] in ("+", "-") else order

        if field_name in order_fns:
            qs = order_fns[field_name](req, qs, direction)
        else:
            field = getattr(qs.model, field_name)
            if direction == "desc":
                field = field.desc()
            qs = qs.order_by(field)
    return qs


def skimmed(req, qs, serializer):
    return serializer(
        qs,
        project=req.params.getall("field") or None,
        omit=req.params.getall("exclude") or None,
    )


def object_list(
    req,
    qs,
    serializer=None,
    key_name=None,
    filter_fns=None,
    order_fns=None,
    paginate_by=15,
    max_paginate_by=100,
):
    serializer = serializer or qs.model.serializer()
    qs = filtered(req, qs, **(filter_fns or {}))
    qs = ordered(req, qs, **(order_fns or {}))
    (qs, pagination) = paginated(
        req, qs, paginate_by=paginate_by, max_paginate_by=max_paginate_by
    )
    key_name = key_name or f"{qs.model.__name__.lower()}_list"
    return {key_name: skimmed(req, qs, serializer), "pagination": pagination}


def object_detail(req, obj, serializer=None, key_name=None):
    serializer = serializer or obj.__class__.serializer()
    key_name = key_name or obj.__class__.__name__.lower()
    return {key_name: skimmed(req, obj, serializer)}



import funcy as fn
import msgpack
from walrus import *
from .settings import settings

__all__ = ("cache_db", "cache", "evict")

is_str = fn.isa(str)


@fn.LazyObject
def cache_db():
    return Walrus(
        host=settings.redis_host,
        port=settings.redis_port,
        db=settings.redis_db
    )


@fn.decorator
def cache(call, *, key=None, evict_keys=None):
    req = call._args[0]
    key = (
        key
        if is_str(key)
        else key(call)
        if callable(key)
        else ":".join(
            [call._func.__name__]
            + fn.lmap(lambda x: f"{x[0]}-{x[1]}", req.params.items())
            + fn.lmap(lambda x: f"{x[0]}-{x[1]}", call._kwargs.items())
        )
    )
    key = f"catalog:cache:{key}"
    resp = cache_db.get(key)
    if resp:
        return msgpack.unpackb(resp, raw=False)

    lock = cache_db.lock(key, ttl=settings.cache_lock_duration)
    with lock:
        resp = call()
        cache_db[key] = msgpack.packb(resp, use_bin_type=True)
        evict_keys = (
            evict_keys
            if fn.is_list(evict_keys)
            else [evict_keys]
            if is_str(evict_keys)
            else evict_keys(call)
            if callable(evict_keys)
            else []
        )
        evict_keys = [ek.format(**call._kwargs) for ek in evict_keys]
        for evict_key in evict_keys:
            cache_set = cache_db.Set(f"catalog:eviction:{evict_key}")
            cache_set.add(key)

    return resp


@fn.decorator
def evict(call, *evict_keys):
    resp = call()
    for evict_key in evict_keys:
        evict_key = evict_key(call) if callable(evict_key) else evict_key
        evict_key = evict_key.format(**call._kwargs)
        cache_set = cache_db.Set(f"catalog:eviction:{evict_key}")
        for key in cache_set.members():
            cache_db.delete(key)
        cache_set.clear()
    return resp



from gevent.local import local
from .http import http

__all__ = ("g", )

g = local()


@http.middleware()
def gevent_local_session_middleware(req, **opts):
    # populates the session object with the incoming request
    g.request = req



from pibe_ext.http import http
from werkzeug.debug import DebuggedApplication


@http.wsgi_middleware()
def install_debugger_middleware(application, **opts):
    if opts.get("use_debugger", False) == True:
        application = DebuggedApplication(application, evalex=True)
    return application



import logging
from .http import http, no_content

logger = logging.getLogger(__name__)

__all__ = ("rpc",)

# https://www.jsonrpc.org/specification

class RPCRegistry(dict):
    def register(self, method_name=None):
        def func_decorator(func):
            self[method_name or func.__name__] = func
            return func

        return func_decorator

    def dispatch(self, wh_name, *args, **kwargs):
        for callback_fn in self.get(wh_name, []):
            callback_fn(*args, **kwargs)


rpc = RPCRegistry()


@http.post("/rpc")
def process_rpc(req):
    payload = req.json
    rpc_method = payload['method']
    rpc_params = payload['params']
    rpc_id = payload['id']

    resp = {
        "jsonrpc": "2.0",
        "id": rpc_id
    }

    try:
        resp["result"] = rpc["method"](*rpc_params)
    except:
        resp["error"] = "An error ocurred"

    return resp



import os
import logging
import funcy as fn
from types import SimpleNamespace

__all__ = ("settings",)

class SettingsRegistry(SimpleNamespace):
    default_fns = []

    def defaults(self):
        def func_decorator(func):
            self.default_fns.append(func)
            return func
        return func_decorator

    def get_defaults(self, **opts):
        exec_fns = [f for f in self.default_fns]
        skip = opts.get("skip", [])
        if skip:
            exec_fns = [f for f in exec_fns if f.__name__ not in skip]

        only = opts.get("only", [])
        if only:
            exec_fns = [f for f in exec_fns if f.__name__ in only]

        return fn.merge(*[f(**opts) for f in exec_fns])

    def initialize(self, **opts):
        self.__dict__.update(self.get_defaults(**opts) or {})

    def get(self, *args):
        return self.__dict__.get(*args)


settings = SettingsRegistry()



import gettext
import funcy as fn
from .session import g
from .http import http, expectation_failed

__all__ = ("get_language", "language_middleware")

@fn.memoize
def get_language(language_code):
    return gettext.translation(
        language_code, localedir="locales/", languages=[language_code]
    )


@http.middleware()
def language_middleware(req, **opts):
    language_code = (
        req.headers.get("X-Language-Code") or req.params.get("language_code") or "en"
    )

    if language_code not in ["en", "pt_PT"]:
        expectation_failed(error="Invalid language requested.")

    req.environ["language_code"] = language_code
    g.language_code = language_code

    if language_code == "en":
        gettext.install("base", "locales")
    else:
        language = get_language(language_code)
        language.install()



class lazy_object(object):
    """
    Create a proxy or placeholder for another object.
    """

    __slots__ = (
        "obj",
        "_callbacks",
        "_init_callback",
        "_check_and_initialize",
    )

    def __init__(self, init_callback=None):
        self._callbacks = []
        self._init_callback = init_callback
        self.initialize(None)

        def _check_and_initialize(self):
            if self.obj is None:
                if self._init_callback:
                    self.initialize(self._init_callback())
                else:
                    raise AttributeError("Cannot use uninitialized Proxy.")

        self._check_and_initialize = _check_and_initialize

    def initialize(self, obj):
        self.obj = obj
        for callback in self._callbacks:
            callback(obj)

    def attach_callback(self, callback):
        self._callbacks.append(callback)
        return callback

    def passthrough(method):
        def inner(self, *args, **kwargs):
            self._check_and_initialize(self)
            return getattr(self.obj, method)(*args, **kwargs)

        return inner

    # Allow proxy to be used as a context-manager.
    __enter__ = passthrough("__enter__")
    __exit__ = passthrough("__exit__")
    __call__ = passthrough("__call__")

    def __getattr__(self, attr):
        if attr not in self.__slots__:
            self._check_and_initialize(self)
        return getattr(self.obj, attr)

    def __setattr__(self, attr, value):
        if attr not in self.__slots__:
            raise AttributeError("Cannot set attribute on proxy.")

        return super(lazy_object, self).__setattr__(attr, value)

    def __getitem__(self, key):
        self._check_and_initialize(self)
        return self.obj.__getitem__(key)

    def __setitem__(self, key, value):
        self._check_and_initialize(self)
        return self.obj.__setitem__(key, value)



import logging
from .http import http, no_content

logger = logging.getLogger(__name__)

__all__ = ("webhook",)


class WebhookRegistry(dict):
    rollback_fns = dict()

    def register(self, wh_name):
        def func_decorator(func):
            if wh_name not in self:
                self[wh_name] = []
            self[wh_name].append(func)
            return func

        return func_decorator

    def rollback(self, wh_name):
        def func_decorator(func):
            if wh_name not in self.rollback_fns:
                self.rollback_fns[wh_name] = []
            self.rollback_fns[wh_name].append(func)
            return func

        return func_decorator

    def __call__(self, wh_name, *args, **kwargs):
        try:
            for func in self.get(wh_name, []):
                func(*args, **kwargs)
        except:
            for rb_func in self.rollback_fns.get(wh_name, []):
                rb_func(*args, **kwargs)
            raise

webhook = WebhookRegistry()



import uuid
from .session import g
from .http import http

__all__ = ("get_correlation_id", )

def get_correlation_id(req):
    return (
        req.headers.get("X-Correlation-ID")
        or req.params.get("correlation_id")
        or str(uuid.uuid4())
    )

@http.middleware()
def correlation_middleware(req, **opts):
    correlation_id = get_correlation_id(req)
    req.environ["correlation_id"] = correlation_id
    g.correlation_id = correlation_id



import sys
import json
import logging
from functools import partial
import funcy as fn

from webob.dec import wsgify

try:
    import peewee as pw
    from playhouse.db_url import connect
    from playhouse.shortcuts import model_to_dict, dict_to_model
    from playhouse.signals import Model as SignalModel
except ImportError:
    raise ImportError("peewee has to be installed to use the db extension")


from .settings import settings
from .crud import model_serializer
# from .init import init
from .http import http



logger = logging.getLogger(__name__)

__all__ = (
    "database",
    "Model",
    "db_models",
    "get_model_class",
    "initialize_database",
    "synchronize_database",
    "database_middleware",
    "db_connect",
    "db_atomic",
)


database = pw.Proxy()


class Model(pw.Model):
    class Meta:
        database = database

    @classmethod
    def serializer(cls, *a, **kw):
        return model_serializer(cls, *a, **kw)

    @classmethod
    def get_or_none(cls, *a, **kw):
        try:
            return cls.get(*a, **kw)
        except cls.DoesNotExist:
            return None

    def refresh(self):
        """To be used in tests"""
        return type(self).get(self._pk_expr())

    def update_from_dict(self, data):
        for key in data:
            if key != "id":
                if getattr(self, key) != data[key]:
                    setattr(self, key, data[key])


@fn.memoize
def db_models():
    return Model.__subclasses__()


@fn.memoize
def get_model_class(class_name):
    for model_class in db_models():
        if class_name in [model_class.__name__, model_class.__name__.lower()]:
            return model_class
    raise ValueError("No model with class {}".format(class_name))


@http.initialize()
def initialize_database(**opts):
    """Initialization method for the database.

    db is a Proxy object for the peewee database. This way it needs to be
    initialized. If connection fails it will issue a system exit.

    """
    db_url = opts.get("db_url")
    db_obj = connect(db_url or settings.database_url)
    database.initialize(db_obj)

    database.connect()
    database.close()
    try:
        database.connect()
        database.close()
    except:
        logger.error("Error connecting to database", exc_info=True)
        sys.exit(1)

    return db_obj


@fn.decorator
def db_connect(call):
    database.connect(reuse_if_open=True)
    try:
        resp = call()
    finally:
        if not database.is_closed():
            database.close()
    return resp


@fn.decorator
def db_atomic(call):
    with database.atomic():
        resp = call()
    return resp


@db_connect
def synchronize_database(drop_tables=None):
    if drop_tables:
        database.drop_tables(drop_tables)
    database.create_tables(db_models(), safe=True)


@wsgify.middleware
def database_middleware(req, app):
    database.connect(reuse_if_open=True)
    try:
        resp = req.get_response(app)
    finally:
        if not database.is_closed():
            database.close()
    return resp


@http.wsgi_middleware()
def add_database_middleware(application, **opts):
    if opts.get("database_middleware", True) == True:
        application = database_middleware(application)
    return application



import funcy as fn
from webob import exc
from json.decoder import JSONDecodeError
import cerberus

from .http import _raise_exc, not_acceptable

__all__ = ("validate", )


@fn.decorator
def validate(
    call,
    schema,
    data_source="json_body",
    exception_class=exc.HTTPBadRequest,
    **kwargs):

    if data_source == "json_body":
        try:
            data = call.req.json
        except JSONDecodeError:
            not_acceptable(error="Invalid JSON Request")

    elif data_source == "params":
        data = call.req.params
    else:
        raise KeyError("unknown data source")

    v = cerberus.Validator(schema, **kwargs)
    if not v.validate(data):
        _raise_exc(exception_class, errors=v.errors)

    call.req.data = v.document

    return call()



import logging
from app.core import settings, http, g

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.wsgi import SentryWsgiMiddleware


@http.initialize()
def init_sentry(**opts):
    if settings.use_sentry:
        sentry_sdk.init(
            dsn=settings.sentry_dsn,
            integrations=[
                LoggingIntegration(
                    level=settings.log_level, event_level=settings.log_level
                )
            ],
        )
        sentry_sdk.set_tag("service", "core-service")


@http.wsgi_middleware()
def install_sentry_middleware(application, **opts):
    if settings.use_sentry:
        application = SentryWsgiMiddleware(application)
    return application


@http.middleware()
def add_correlation_id(req, **opts):
    if settings.use_sentry:
        sentry_sdk.set_tag("correlation_id", g.correlation_id)
